# sage_setup: distribution = sagemath-brial
# sage.doctest: needs sage.rings.polynomial.pbori
from sage.rings.polynomial.pbori.pbori import (
    BooleConstant,
    Monomial,
    Polynomial,
    top_index,
)


def used_vars(l, bound=None):
    if not l:
        return BooleConstant(1)
    m = Monomial(Polynomial(next(iter(l))).vars_as_monomial())
    for p in l[1:]:
        m = m * Polynomial(p).vars_as_monomial()
        if bound and len(m) > bound:
            return m
    return m


def used_vars_set(l, bound=None):
    if not l:
        return BooleConstant(1)
    s = set()
    for p in l:
        s.update(Polynomial(p).vars_as_monomial().variables())
        if bound and len(s) > bound:
            break
    sorted_s = sorted(s, key=top_index, reverse=True)
    m = Monomial(next(iter(l)).ring())
    for v in sorted_s:
        m = v * m

    return m



# sage_setup: distribution = sagemath-brial
# sage.doctest: needs sage.rings.polynomial.pbori
from pprint import pformat
from random import Random

from sage.rings.polynomial.pbori.blocks import declare_ring
from sage.rings.polynomial.pbori.ll import ll_encode
from sage.rings.polynomial.pbori.pbori import (
    Monomial,
    Polynomial,
    Variable,
    ll_red_nf_redsb,
    random_set,
    set_random_seed,
)


def gen_random_poly(ring, l, deg, vars_set, seed=123):
    """
    Generate a random polynomial with coefficients in ``ring``.

    EXAMPLES::

        sage: from sage.rings.polynomial.pbori.PyPolyBoRi import Ring, Variable
        sage: from sage.rings.polynomial.pbori.randompoly import gen_random_poly
        sage: r = Ring(16)
        sage: vars = [Variable(i,r) for i in range(10)]
        sage: gen_random_poly(r, 4, 10, vars)  # random
        x(0)*x(1)*x(2)*x(5)*x(8)*x(9) + x(0)*x(1)*x(4)*x(6) + x(0)*x(2)*x(3)*x(7)*x(9) + x(5)*x(8)
    """
    myrange = vars_set
    r = Random(seed)

    def helper(samples):
        if samples == 0:
            return Polynomial(ring.zero())
        if samples == 1:
            d = r.randint(0, deg)
            variables = r.sample(myrange, d)
            m = Monomial(ring)
            for v in sorted(set(variables), reverse=True):
                m = m * Variable(v, ring)
            return Polynomial(m)
        assert samples >= 2
        return helper(samples // 2) + helper(samples - samples // 2)
    p = Polynomial(ring.zero())
    while len(p) < l:
        p = Polynomial(p.set().union(helper(l - len(p)).set()))
    return p


def sparse_random_system(ring, number_of_polynomials, variables_per_polynomial,
                         degree, random_seed=None):
    r"""
    Generate a sparse random system.

    Generate a system, which is sparse in the sense, that each polynomial
    contains only a small subset of variables. In each variable that occurs
    in a polynomial it is dense in the terms up to the given degree
    (every term occurs with probability 1/2).

    The system will be satisfiable by at least one solution.

    TESTS::

        sage: from sage.rings.polynomial.pbori import Ring, groebner_basis
        sage: r = Ring(10)
        sage: from sage.rings.polynomial.pbori.randompoly import sparse_random_system
        sage: s = sparse_random_system(r, number_of_polynomials=20, variables_per_polynomial=3, degree=2, random_seed=int(123))
        sage: [p.deg() for p in s]
        [2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2]
        sage: sorted(groebner_basis(s), reverse=True)
        [x(0), x(1) + 1, x(2), x(3) + 1, x(4) + 1, x(5), x(6), x(7) + 1, x(8) + 1, x(9) + 1]
    """
    if random_seed is not None:
        set_random_seed(random_seed)
    random_generator = Random(random_seed)
    variables = [ring.variable(i) for i in range(ring.n_variables())]
    solutions = [v + random_generator.randint(0, 1) for v in variables]
    solutions = ll_encode(solutions)
    res = []
    while len(res) < number_of_polynomials:
        variables_as_monomial = Monomial(
            random_generator.sample(
                variables,
                variables_per_polynomial)
        )
        p = Polynomial(random_set(variables_as_monomial, 2 ** (
            variables_per_polynomial - 1)))
        p = sum([p.graded_part(i) for i in range(degree + 1)])
        if p.deg() == degree:
            res.append(p)
    # evaluate it to guarantee a solution
    return [p + ll_red_nf_redsb(p, solutions) for p in res]


def sparse_random_system_data_file_content(number_of_variables, **kwds):
    r"""
    TESTS::

        sage: from sage.rings.polynomial.pbori.randompoly import sparse_random_system_data_file_content
        sage: sparse_random_system_data_file_content(10, number_of_polynomials=5, variables_per_polynomial=3, degree=2, random_seed=int(123))
        "declare_ring(['x'+str(i) for in range(10)])\nideal=\\\n[...]\n\n"
    """
    dummy_dict = {}
    r = declare_ring(['x' + str(i) for i in range(number_of_variables)],
                     dummy_dict)
    polynomials = sparse_random_system(r, **kwds)
    polynomials = pformat(polynomials)
    return "declare_ring(['x'+str(i) for in range({})])\nideal=\\\n{}\n\n".format(
        number_of_variables, polynomials)



# sage_setup: distribution = sagemath-brial
# sage.doctest: needs sage.rings.polynomial.pbori
from sage.rings.polynomial.pbori.pbori import (
    BooleConstant,
    BooleSet,
    Monomial,
    Polynomial,
    Variable,
    if_then_else,
    mod_mon_set,
    top_index,
)


def all_monomials_of_degree_d_old(d, variables):
    """
    Return monomials of degree d in the given variables.

    Obsolete version ?
    """
    if d == 0:
        return BooleConstant(1)

    if not variables:
        return []
    variables = sorted(set(variables), reverse=True, key=top_index)

    m = variables[-1]
    for v in variables[:-1]:
        m = v + m
    m = m.set()
    i = 0
    res = Polynomial(variables[0].ring().one()).set()
    while i < d:
        i += 1
        res = res.cartesian_product(m).diff(res)
    return res


def all_monomials_of_degree_d(d, variables):
    """
    Return monomials of degree d in the given variables.
    """
    variables = Monomial(variables)
    variables = list(variables.variables())
    if not variables:
        assert d == 0
        return BooleConstant(1)
    ring = variables[0].ring()
    if d > len(variables):
        return Polynomial(0, ring)
    if d < 0:
        return Polynomial(1, ring)

    deg_variables = variables[-d:]
    # this ensures sorting by indices
    res = Monomial(deg_variables)

    for i in range(1, len(variables) - d + 1):
        deg_variables = variables[-d - i:-i]
        res = Polynomial(res)
        nav = res.navigation()
        navs = []
        while not nav.constant():
            navs.append(BooleSet(nav, ring))
            nav = nav.then_branch()
        acc = Polynomial(1, ring)
        for (nav, v) in reversed(zip(navs, deg_variables)):
            acc = if_then_else(v, acc, nav)
        res = acc
    return res.set()


def power_set(variables):
    """
    Return all subsets of the given variables.
    """
    if not variables:
        return BooleConstant(1)
    variables = sorted(set(variables), reverse=True, key=top_index)
    res = Polynomial(1, variables[0].ring()).set()
    for v in variables:
        res = if_then_else(v, res, res)
    return res


if __name__ == '__main__':
    from .blocks import Block, declare_ring
    r = declare_ring([Block("x", 10000)], globals())
    print(list(all_monomials_of_degree_d(0, [Variable(i) for i in range(100)])))
    print(list(all_monomials_of_degree_d(1, [Variable(i) for i in range(10)])))
    print(list(all_monomials_of_degree_d(2, [Variable(i) for i in range(4)])))
    print(list(all_monomials_of_degree_d(3, [Variable(i) for i in range(4)])))
    print(list(all_monomials_of_degree_d(4, [Variable(i) for i in range(4)])))
    print(list(all_monomials_of_degree_d(0, [])))
    print(list(all_monomials_of_degree_d(1, [])))
    print(list(power_set([Variable(i) for i in range(2)])))
    print(list(power_set([Variable(i) for i in range(4)])))
    print(list(power_set()))
    # every monomial in the first 8 var, which is at most linear in the first 5
    print(list(mod_mon_set(
        power_set([Variable(i) for i in range(8)]),
        all_monomials_of_degree_d(2, [Variable(i) for i in range(5)]))))

    # specialized normal form computation
    print(Polynomial(
        mod_mon_set(
            (x(1) * x(2) + x(1) + 1).set(),
            all_monomials_of_degree_d(2, [Variable(i) for i in range(1000)]))))
    print(list(mod_mon_set(
        power_set([Variable(i) for i in range(50)]),
        all_monomials_of_degree_d(2, [Variable(i) for i in range(1000)]))))


def monomial_from_indices(ring, indices):
    res = Monomial(ring)
    for i in sorted(indices, reverse=True):
        res = res * ring.variable(i)
    return res



# sage_setup: distribution = sagemath-brial
"""
PolyBoRi package: Polynomials in Boolean Ring

The PolyBoRi package implements a framework for computations with Polynomials in Boolean Ring.

The core of PolyBoRi is a C++ library, which provides high-level data types for Boolean polynomials and monomials,
exponent vectors, as well as for the underlying polynomial rings and subsets of the powerset of the Boolean variables.
The description of the latter can be found in the description of the 'dynamic' submodule, as well as in the doxygen-based documentation.

As a unique approach, binary decision diagrams are used as internal storage type for polynomial structures.
On top of this C++-library we provide a Python interface. This allows parsing of complex polynomial systems,
as well as sophisticated and extendable strategies for Groebner base computation.
PolyBoRi features a powerful reference implementation for Groebner basis computation.

AUTHOR:

The PolyBoRi Team, 2007-2011

REFERENCES:

M. Brickenstein, A. Dreyer, G. Greuel, M. Wedler, O. Wienand,
*New developments in the theory of Groebner bases and applications
to formal Verification*,  Preprint at :arxiv:`0801.1177`

M. Brickenstein, A. Dreyer, PolyBoRi:
*A Groebner Basis Framework for Boolean Polynomials*,
Reports of Fraunhofer ITWM, No. 122, Kaiserslautern, Germany, 2007.
http://www.itwm.fraunhofer.de/zentral/download/berichte/bericht122.pdf

M. Brickenstein, A. Dreyer, PolyBoRi:
*A framework for Groebner basis computations with Boolean polynomials*,
Electronic Proceedings of the MEGA 2007 - Effective Methods in Algebraic Geometry, Strobl, Austria, June 2007.
http://www.ricam.oeaw.ac.at/mega2007/electronic/electronic.html
"""
from .PyPolyBoRi import Ring, Polynomial, Monomial, Variable

# Get all-inclusive groebner routine
from .gbcore import groebner_basis
from .nf import normal_form

# Import some high-level modelling functionality
from .blocks import declare_ring
from .blocks import HigherOrderBlock, AlternatingBlock, Block
from .gbrefs import load_file
from .specialsets import all_monomials_of_degree_d, power_set



#!/usr/bin/env python

# PEP 517 builds do not have . in sys.path
import os
import sys
sys.path.insert(0, os.path.dirname(__file__))

from sage_setup import sage_setup

sage_setup(['sagemath-brial'],
           recurse_packages=('sage', 'passagemath_brial'),
           required_modules=('m4ri', 'libpng'),
           spkgs=['brial', 'm4ri', 'libpng', 'gmp', 'mpc', 'mpfr'],
           package_data={})



# sage_setup: distribution = sagemath-brial
# sage.doctest: needs sage.rings.polynomial.pbori
from pathlib import Path
from warnings import warn

from sage.rings.polynomial.pbori.easy_polynomials import (
    easy_linear_polynomials as easy_linear_polynomials_func,
)
from sage.rings.polynomial.pbori.pbori import (
    BooleSet,
    GroebnerStrategy,
    Monomial,
    Polynomial,
    ReductionStrategy,
    Variable,
    easy_linear_factors,
    mod_mon_set,
    parallel_reduce,
)
from sage.rings.polynomial.pbori.PyPolyBoRi import (
    BoolePolynomialVector,
)
from sage.rings.polynomial.pbori.statistics import used_vars_set


class GeneratorLimitExceeded(Exception):
    r"""
    Docstring for GeneratorLimitExceeded
    """

    def __init__(self, strat):
        self.strat = strat


def pkey(p):
    return (p[0], len(p))


mat_counter = 0


def build_and_print_matrices(v, strat):
    r"""
    Old solution using PIL, the currently used implementation is done in C++
    and plots the same matrices, as being calculated
    """
    treated = BooleSet()
    v = list(v)
    if not v:
        return
    rows = 0
    polys_in_mat = []
    while v:
        rows = rows + 1
        p = v[0]
        v = v[1:]
        for m in list(p.terms()):
            mom = Monomial(m)
            if mom not in BooleSet(treated):
                i = strat.select(mom)
                if i >= 0:
                    p2 = strat[i]
                    p2 = p2 * (mom // p2.lead())
                    v.append(p2)
        polys_in_mat.append(p)
        treated = treated.union(p.set())
    m2i = {v: k
           for k, v in enumerate(list(Polynomial(BooleSet(treated)).terms()))}
    polys_in_mat.sort(key=Polynomial.lead, reverse=True)
    polys_in_mat = [[m2i[t] for t in p.terms()] for p in polys_in_mat]

    global mat_counter
    mat_counter = mat_counter + 1
    from PIL import Image

    rows = len(polys_in_mat)
    cols = len(m2i)
    im = Image.new("1", (cols, rows), "white")
    for i, pi in enumerate(polys_in_mat):
        for j in pi:
            assert i < rows
            assert j < cols
            im.putpixel((j, i), 0)

    file_name = Path(strat.matrix_prefix + str(mat_counter) + ".png")
    file_name.unlink(missing_ok=True)
    im.save(file_name)
    del im

    print("MATRIX_SIZE:", rows, "x", cols)


def multiply_polynomials(l, ring):
    r"""

    TESTS::

        sage: from sage.rings.polynomial.pbori import *
        sage: r = Ring(1000)
        sage: x = r.variable
        sage: from sage.rings.polynomial.pbori.nf import multiply_polynomials
        sage: multiply_polynomials([x(3), x(2)+x(5)*x(6), x(0), x(0)+1], r)
        0
    """
    l = [Polynomial(p) for p in l]

    def sort_key(p):
        return p.navigation().value()
    l = sorted(l, key=sort_key)
    res = Polynomial(ring.one())
    for p in l:
        res = p * res
    return res


def build_and_print_matrices_deg_colored(v, strat):
    r"""
    old PIL solution using a different color for each degree
    """
    if not v:
        return

    treated = BooleSet()
    v = list(v)
    rows = 0
    polys_in_mat = []

    while v:
        rows = rows + 1
        p = v[0]
        v = v[1:]
        for m in list(p.terms()):
            mom = Monomial(m)
            if m not in BooleSet(treated):
                i = strat.select(mom)
                if i >= 0:
                    p2 = strat[i]
                    p2 = p2 * (mom // p2.lead())
                    v.append(p2)
        polys_in_mat.append(p)
        treated = treated.union(p.set())
    m2i = {v: k for k, v in enumerate(BooleSet(treated))}
    max_deg = max(m.deg() for m in BooleSet(treated))
    if max_deg == 0:
        max_deg = 1
    i2deg = {m2i[m]: m.deg() for m in BooleSet(treated)}
    polys_in_mat = [[m2i[t] for t in p.terms()] for p in polys_in_mat]
    polys_in_mat.sort(key=pkey)
    global mat_counter
    mat_counter = mat_counter + 1
    from PIL import Image, ImageColor

    rows = len(polys_in_mat)
    cols = len(m2i)
    im = Image.new("RGB", (cols, rows), "white")
    for i in range(len(polys_in_mat)):
        p = polys_in_mat[i]
        for j in p:
            assert i < rows
            assert j < cols
            hsl = str(270 - (270 * i2deg[j]) / max_deg)
            im.putpixel((j, i), ImageColor.getrgb("hsl(" + hsl + ",100%,50%)"))
    file_name = Path(strat.matrix_prefix + str(mat_counter) + ".png")
    file_name.unlink(missing_ok=True)
    im.save(file_name)
    del im

    print("MATRIX_SIZE:", rows, "x", cols)


def high_probability_polynomials_trick(p, strat):
    lead_deg = p.lead_deg()
    if lead_deg <= 4:
        return

    ring = p.ring()
    factor = multiply_polynomials(easy_linear_factors(p), ring)
    p = p / factor

    # again, do it twice, it's cheap
    lead_deg = p.lead_deg()
    if lead_deg <= 3:
        return

    if lead_deg > 9:
        return
    uv = p.vars_as_monomial()

    candidates = []

    if uv.deg() <= 4:
        return

    if not uv.deg() <= lead_deg + 1:
        return

    space = uv.divisors()

    lead = p.lead()
    for v in lead.variables():
        variable_selection = lead // v
        vars_reversed = list(reversed(variable_selection.variables()))
        # it's just a way to loop over the cartesian product
        for assignment in variable_selection.divisors():
            c_p = assignment
            for var in vars_reversed:
                if not assignment.reducible_by(var):
                    c_p = (var + 1) * c_p

            points = (c_p + 1).zeros_in(space)
            if p.zeros_in(points).empty():
                candidates.append(c_p * factor)
        # there many more combinations depending on plugged in values
    for c in candidates:
        strat.add_as_you_wish(c)


def symmGB_F2_python(G, deg_bound=1000000000000, over_deg_bound=0,
                     use_faugere=False, use_noro=False,
                     opt_lazy=True, opt_red_tail=True,
                     max_growth=2.0, step_factor=1.0,
                     implications=False, prot=False,
                     full_prot=False, selection_size=1000, opt_exchange=True,
                     opt_allow_recursion=False, ll=False,
                     opt_linear_algebra_in_last_block=True,
                     max_generators=None,
                     red_tail_deg_growth=True, matrix_prefix='mat',
                     modified_linear_algebra=True, draw_matrices=False,
                     easy_linear_polynomials=True):
    if use_noro and use_faugere:
        raise ValueError('both use_noro and use_faugere specified')

    def add_to_basis(strat, p):
        if p.is_zero():
            if prot:
                print("-")
        else:
            if prot:
                if full_prot:
                    print(p)
                print("Result: ", "deg:", p.deg(), "lm: ",
                      p.lead(), "el: ", p.elength())
            if easy_linear_polynomials and p.lead_deg() > 2:
                lin = easy_linear_polynomials_func(p)
                for q in lin:
                    strat.add_generator_delayed(q)
            old_len = len(strat)
            strat.add_as_you_wish(p)
            new_len = len(strat)
            if new_len == 1 + old_len:
                high_probability_polynomials_trick(p, strat)

            if prot:
                print("#Generators:", len(strat))

    if isinstance(G, list):
        if not G:
            return []
        G = [Polynomial(g) for g in G]
        current_ring = G[0].ring()
        strat = GroebnerStrategy(current_ring)
        strat.reduction_strategy.opt_red_tail = opt_red_tail
        strat.opt_lazy = opt_lazy
        strat.opt_exchange = opt_exchange
        strat.opt_allow_recursion = opt_allow_recursion
        strat.enabled_log = prot
        strat.reduction_strategy.opt_ll = ll
        strat.opt_modified_linear_algebra = modified_linear_algebra
        strat.opt_linear_algebra_in_last_block = (
            opt_linear_algebra_in_last_block)
        strat.opt_red_by_reduced = False  # True
        strat.reduction_strategy.opt_red_tail_deg_growth = red_tail_deg_growth

        strat.opt_draw_matrices = draw_matrices
        strat.matrix_prefix = matrix_prefix

        for g in G:
            if not g.is_zero():
                strat.add_generator_delayed(g)
    else:
        strat = G

    if prot:
        print("added delayed")
    i = 0
    try:
        while strat.npairs() > 0:
            if max_generators and len(strat) > max_generators:
                raise GeneratorLimitExceeded(strat)
            i = i + 1
            if prot:
                print("Current Degree:", strat.top_sugar())
            if (strat.top_sugar() > deg_bound) and (over_deg_bound <= 0):
                return strat
            if (strat.top_sugar() > deg_bound):
                ps = strat.some_spolys_in_next_degree(over_deg_bound)
                over_deg_bound -= len(ps)
            else:
                ps = strat.some_spolys_in_next_degree(selection_size)

            if ps and ps[0].ring().has_degree_order():
                ps = [strat.reduction_strategy.cheap_reductions(p) for p in ps]
                ps = [p for p in ps if not p.is_zero()]
                if ps:
                    min_deg = min(p.deg() for p in ps)
                new_ps = []
                for p in ps:
                    if p.deg() <= min_deg:
                        new_ps.append(p)
                    else:
                        strat.add_generator_delayed(p)
                ps = new_ps

            if prot:
                print("(", strat.npairs(), ")")
            if prot:
                print("start reducing")
                print("Chain Crit. : ", strat.chain_criterions, "VC:", strat.
                      variable_chain_criterions, "EASYP", strat.
                      easy_product_criterions, "EXTP", strat.
                      extended_product_criterions)
                print(len(ps), "spolys added")

            if use_noro or use_faugere:
                v = BoolePolynomialVector()

                for p in ps:
                    if not p.is_zero():
                        v.append(p)
                res = strat.noro_step(v) if use_noro else strat.faugere_step_dense(v)

            else:
                v = BoolePolynomialVector()
                for p in ps:
                    rp = Polynomial(mod_mon_set(
                        BooleSet(p.set()),
                        strat.reduction_strategy.monomials))
                    if not rp.is_zero():
                        v.append(rp)
                if len(v) > 100:
                    res = parallel_reduce(v, strat, int(step_factor * 10),
                                          max_growth)
                elif len(v) > 10:
                    res = parallel_reduce(v, strat, int(step_factor * 30),
                                          max_growth)
                else:
                    res = parallel_reduce(v, strat, int(step_factor * 100),
                                          max_growth)

            if prot:
                print("end reducing")

            if res and res[0].ring().has_degree_order():
                res_min_deg = min([p.deg() for p in res])
                new_res = []
                for p in res:
                    if p.deg() == res_min_deg:
                        new_res.append(p)
                    else:
                        strat.add_generator_delayed(p)
                res = new_res

            def sort_key(p):
                return p.lead()
            res_cp = sorted(res, key=sort_key)

            for p in res_cp:
                old_len = len(strat)
                add_to_basis(strat, p)
                if implications and old_len == len(strat) - 1:
                    strat.implications(len(strat) - 1)
                if p.is_one():
                    if prot:
                        print("GB is 1")
                    return strat
                if prot:
                    print("(", strat.npairs(), ")")

            strat.clean_top_by_chain_criterion()
        return strat
    except KeyboardInterrupt:
        raise


def GPS(G, vars_start, vars_end):
    def step(strat, trace, var, val):
        print("plugin: ", var, val)
        print("npairs", strat.npairs())
        strat = GroebnerStrategy(strat)
        print("npairs", strat.npairs())
        strat.add_generator_delayed(Polynomial(
            Monomial(Variable(var, strat.r)) + val))
        strat = symmGB_F2_python(strat, prot=True, deg_bound=2,
                                 over_deg_bound=10)
        if var <= vars_start:
            strat = symmGB_F2_python(strat, prot=True, opt_lazy=False,
                                     opt_red_tail=False)
        if strat.containsOne():
            pass
        else:
            if var <= vars_start:
                # bug: may contain Delayed polynomials
                print("!!!!!!! SOLUTION", trace)
                raise Exception
                # yield trace
            branch(strat, trace + [(var, val)], var - 1)

    def branch(strat, trace, var):
        while strat.variableHasValue(var):
            # remember to add value to trace
            var -= 1
        step(strat, trace, var, 0)
        step(strat, trace, var, 1)
    if G:
        strat = GroebnerStrategy(G[0].ring())
        # strat.add_generator(G[0])
        for g in G[:]:
            strat.add_generator_delayed(g)
        branch(strat, [], vars_end - 1)


def GPS_with_proof_path(G, proof_path, deg_bound, over_deg_bound):
    def step(strat, trace, proof_path, pos, val):
        print(proof_path)
        print("plugin: ", pos, val, proof_path[pos])
        print("npairs", strat.npairs())
        strat = GroebnerStrategy(strat)
        print("npairs", strat.npairs())
        print("npairs", strat.npairs())
        plug_p = Polynomial(proof_path[pos]) + val
        plug_p_lead = plug_p.lead()
        if len(plug_p) == 2 and (plug_p + plug_p_lead).deg() == 0:
            for v in plug_p_lead:
                strat.add_generator_delayed(v + 1)
        else:
            strat.add_generator_delayed(plug_p)
        print("npairs", strat.npairs())
        print("pos:", pos)
        strat = symmGB_F2_python(strat, deg_bound=deg_bound, opt_lazy=False,
                                 over_deg_bound=over_deg_bound, prot=True)
        print("npairs", strat.npairs())
        pos = pos + 1
        if pos >= len(proof_path):
            print("OVER")
            strat = symmGB_F2_python(strat, prot=True)
        if strat.containsOne():
            pass
        else:
            if pos >= len(proof_path):
                print("npairs", strat.npairs())
                print("minimized:")
                for p in strat.minimalize_and_tail_reduce():
                    print(p)
                # bug: may contain Delayed polynomials
                print("!!!!!!! SOLUTION", trace)
                raise Exception
                # yield trace
            branch(strat, trace + [(pos, val)], proof_path, pos)

    def branch(strat, trace, proof_path, pos):

        step(strat, trace, proof_path, pos, 0)
        step(strat, trace, proof_path, pos, 1)
    strat = GroebnerStrategy(G[0].ring())
    strat.add_generator(Polynomial(G[0]))
    for g in G[1:]:
        strat.add_generator_delayed(Polynomial(g))
    branch(strat, [], proof_path, 0)


def GPS_with_suggestions(G, deg_bound, over_deg_bound, opt_lazy=True,
                         opt_red_tail=True, initial_bb=True):
    def step(strat, trace, var, val):
        print(trace)
        plug_p = val + var
        print("plugin: ", len(trace), plug_p)
        trace = trace + [plug_p]
        strat = GroebnerStrategy(strat)

        strat.add_generator_delayed(plug_p)
        print("npairs", strat.npairs())

        strat = symmGB_F2_python(strat, deg_bound=deg_bound,
                                 opt_lazy=opt_lazy,
                                 over_deg_bound=over_deg_bound, prot=True)

        if not strat.containsOne():
            branch(strat, trace)

    def branch(strat, trace):
        print("branching")
        index = strat.suggestPluginVariable()

        if index < 0:
            uv = set(used_vars_set(strat))
            lv = {next(iter(p.lead())).index()
                  for p in strat if p.lead_deg() == 1}
            candidates = uv.difference(lv)
            if candidates:
                index = next(iter(candidates)).index()
        if index >= 0:
            print("chosen index:", index)
            step(strat, trace, Polynomial(Monomial(Variable(index))), 0)
            step(strat, trace, Polynomial(Monomial(Variable(index))), 1)
        else:
            print("FINAL!!!", index)
            strat = symmGB_F2_python(strat, prot=True)
            if not strat.containsOne():
                print("TRACE", trace)
                print("SOLUTION")
                for p in strat.minimalize_and_tail_reduce():
                    print(p)
                raise Exception

    def sort_crit(p):
        return (p.lead(), p.deg(), p.elength())
    if not G:
        return
    strat = GroebnerStrategy(G[0].ring())
    strat.reduction_strategy.opt_red_tail = opt_red_tail  # True
    strat.opt_exchange = False
    strat.opt_allow_recursion = False
    strat.opt_lazy = opt_lazy
    first_deg_bound = 1
    G = [Polynomial(p) for p in G]
    G.sort(key=sort_crit)
    if initial_bb:
        for g in G:
            print(g)
            if g.deg() == 1:  # (index<0):
                strat.add_as_you_wish(g)
            else:
                first_deg_bound = max(first_deg_bound, g.deg())
                strat.add_generator_delayed(g)
            print(g, len(strat))
    else:
        for g in G:
            strat.add_as_you_wish(g)
    if initial_bb:
        strat = symmGB_F2_python(strat, deg_bound=max(deg_bound,
                                                      first_deg_bound),
                                 opt_lazy=opt_lazy, over_deg_bound=0,
                                 prot=True)
    strat.opt_lazy = opt_lazy
    print("INITIALIZED")
    branch(strat, [])


def GPS_with_non_binary_proof_path(G, proof_path, deg_bound, over_deg_bound):
    def step(strat, trace, proof_path, pos, choice):
        print(proof_path)
        print("plugin: ", pos)
        print("npairs", strat.npairs())
        strat = GroebnerStrategy(strat)
        print("npairs", strat.npairs())
        print("npairs", strat.npairs())
        for p in proof_path[pos][choice]:
            print(p)
            strat.add_generator_delayed(Polynomial(p))

        print("npairs", strat.npairs())
        print("pos:", pos)
        strat = symmGB_F2_python(strat, deg_bound=deg_bound,
                                 over_deg_bound=over_deg_bound, prot=True)
        print("npairs", strat.npairs())
        pos = pos + 1
        if pos >= len(proof_path):
            print("OVER")
            strat = symmGB_F2_python(strat)
        if strat.containsOne():
            pass
        else:
            if pos >= len(proof_path):
                print("npairs", strat.npairs())
                # bug: may contain Delayed polynomials
                print("!!!!!!! SOLUTION", trace)
                raise Exception
                # yield trace
            branch(strat, trace + [(pos, choice)], proof_path, pos)
            # workaround because of stack depth
            # step(strat,trace+[(var,val)],var-1, 0)
            # step(strat,trace+[(var,val)],var-1, 1)

    def branch(strat, trace, proof_path, pos):
        for i in range(len(proof_path[pos])):
            step(strat, trace, proof_path, pos, i)

    strat = GroebnerStrategy(G[0].ring())
    strat.add_generator(G[0])
    for g in G[1:]:
        strat.add_generator_delayed(g)
    branch(strat, [], proof_path, 0)


def symmGB_F2_C(G, opt_exchange=True,
                deg_bound=1000000000000, opt_lazy=False,
                over_deg_bound=0, opt_red_tail=True,
                max_growth=2.0, step_factor=1.0,
                implications=False, prot=False,
                full_prot=False, selection_size=1000,
                opt_allow_recursion=False, use_noro=False, use_faugere=False,
                ll=False, opt_linear_algebra_in_last_block=True,
                max_generators=None, red_tail_deg_growth=True,
                modified_linear_algebra=True, matrix_prefix='',
                draw_matrices=False):
    if use_noro:
        raise NotImplementedError("noro not implemented for symmgb")
    if isinstance(G, list):
        if not G:
            return []

        G = [Polynomial(g) for g in G]
        strat = GroebnerStrategy(G[0].ring())
        strat.reduction_strategy.opt_red_tail = opt_red_tail
        strat.enabled_log = prot
        strat.opt_lazy = opt_lazy
        strat.opt_exchange = opt_exchange
        strat.reduction_strategy.opt_ll = ll
        strat.opt_allow_recursion = opt_allow_recursion
        strat.opt_linear_algebra_in_last_block = (
            opt_linear_algebra_in_last_block)
        strat.enabled_log = prot
        strat.opt_modified_linear_algebra = modified_linear_algebra
        strat.matrix_prefix = matrix_prefix
        strat.opt_draw_matrices = draw_matrices
        strat.reduction_strategy.opt_red_tail_deg_growth = red_tail_deg_growth

        strat.redByReduced = False  # True

        for g in G:  # [1:]:
            if not g.is_zero():
                strat.add_generator_delayed(g)
    strat.symmGB_F2()
    return strat


def normal_form(poly, ideal, reduced=True):
    r"""
    Simple normal form computation of a polynomial  against an ideal.

    TESTS::

        sage: from sage.rings.polynomial.pbori import declare_ring, normal_form
        sage: r=declare_ring(['x','y'], globals())
        sage: normal_form(x+y, [y],reduced=True)
        x
        sage: normal_form(x+y,[x,y])
        0
    """
    ring = poly.ring()
    strat = ReductionStrategy(ring)
    strat.opt_red_tail = reduced
    ideal = [Polynomial(p) for p in ideal if p != 0]
    ideal = sorted(ideal, key=Polynomial.lead)
    last = None
    for p in ideal:
        if p.lead() != last:
            strat.add_generator(p)
        else:
            warn("%s will not used for reductions" % p)
        last = p.lead()
    return strat.nf(poly)


def _test():
    import doctest
    doctest.testmod()


if __name__ == "__main__":
    _test()



# sage_setup: distribution = sagemath-brial
# sage.doctest: needs sage.rings.polynomial.pbori
r"""
parallel.py
PolyBoRi

Created by Michael Brickenstein on 2008-10-31.
Copyright 2008 The PolyBoRi Team
"""
import copyreg
import os
from zlib import compress, decompress

from sage.rings.polynomial.pbori.gbcore import groebner_basis
from sage.rings.polynomial.pbori.pbori import (
    BooleSet,
    CCuddNavigator,
    Monomial,
    Polynomial,
    Variable,
    if_then_else,
)
from sage.rings.polynomial.pbori.PyPolyBoRi import (
    Ring,
    WeakRingRef,
)


def to_fast_pickable(l) -> list:
    r"""
    Convert a list of polynomials into a builtin Python value,
    which is fast pickable and compact.

    INPUT:

    - ``l`` -- a list of Boolean polynomials

    OUTPUT:

    It is converted to a tuple consisting of
    - codes referring to the polynomials
    - list of conversions of nodes.
    The nodes are sorted, so that n occurs before n.else_branch(), n.then_branch()
    nodes are only listed, if they are not constant.

    A node is converted in this way:
    0 -> 0
    1 -> 1
    if_then_else(v,t,e) -> (v, index of then branch +2, index of else branch +2)
    the shift of +2 is for the constant values implicitly contained in the list.
    Each code c refers to the c-2-th position in the conversion list, if c >=2, else to
    the corresponding Boolean constant if c in {0, 1}

    EXAMPLES::

        sage: from sage.rings.polynomial.pbori import Ring, Polynomial
        sage: from sage.rings.polynomial.pbori.parallel import to_fast_pickable, from_fast_pickable
        sage: r = Ring(1000)
        sage: x = r.variable
        sage: to_fast_pickable([Polynomial(1, r)])
        [[1], []]
        sage: to_fast_pickable([Polynomial(0, r)])
        [[0], []]
        sage: to_fast_pickable([x(0)])
        [[2], [(0, 1, 0)]]
        sage: to_fast_pickable([x(0)*x(1)+x(1)])
        [[2], [(0, 3, 3), (1, 1, 0)]]
        sage: to_fast_pickable([x(1)])
        [[2], [(1, 1, 0)]]
        sage: to_fast_pickable([x(0)+1])
        [[2], [(0, 1, 1)]]
        sage: to_fast_pickable([x(0)*x(1)])
        [[2], [(0, 3, 0), (1, 1, 0)]]
        sage: to_fast_pickable([x(0)*x(1)+x(1)])
        [[2], [(0, 3, 3), (1, 1, 0)]]
        sage: to_fast_pickable([x(0)*x(1)+x(2)])
        [[2], [(0, 3, 4), (1, 1, 0), (2, 1, 0)]]
        sage: p=x(5)*x(23) + x(5)*x(24)*x(59) + x(5) + x(6)*x(23)*x(89) + x(6)*x(60)*x(89) + x(23) + x(24)*x(89) + x(24) + x(60)*x(89) + x(89) + 1
        sage: from_fast_pickable(to_fast_pickable([p]), r)==[p]
        True
        sage: to_fast_pickable([x(0)*x(1), Polynomial(0, r), Polynomial(1, r), x(3)])
        [[2, 0, 1, 4], [(0, 3, 0), (1, 1, 0), (3, 1, 0)]]
    """
    if not l:
        return [[], []]

    f = l[0]
    f = f.set()
    r = f.ring()
    one = r.one().navigation()
    zero = r.zero().navigation()
    nodes = set()

    def find_navs(nav):
        if nav not in nodes and not nav.constant():
            nodes.add(nav)
            find_navs(nav.then_branch())
            find_navs(nav.else_branch())
    for f in l:
        f_nav = f.set().navigation()
        find_navs(f_nav)

    nodes_sorted = sorted(nodes, key=CCuddNavigator.value)
    nodes2i = {one: 1, zero: 0}
    for i, n in enumerate(nodes_sorted):
        nodes2i[n] = i + 2

    for i in range(len(nodes_sorted)):
        n = nodes_sorted[i]
        t = nodes2i[n.then_branch()]
        e = nodes2i[n.else_branch()]
        nodes_sorted[i] = (n.value(), t, e)

    return [[nodes2i[f.set().navigation()] for f in l], nodes_sorted]


def from_fast_pickable(l, r) -> list:
    r"""
    Undo the operation :func:`to_fast_pickable`.

    The first argument is an object created by :func:`to_fast_pickable`.

    For the specified format, see the documentation of :func:`to_fast_pickable`.
    The second argument is ring, in which this polynomial should be created.

    INPUT:

    See OUTPUT of :func:`to_fast_pickable`.

    OUTPUT: list of Boolean polynomials

    EXAMPLES::

        sage: from sage.rings.polynomial.pbori import Ring
        sage: from sage.rings.polynomial.pbori.parallel import from_fast_pickable
        sage: r = Ring(1000)
        sage: x = r.variable
        sage: from_fast_pickable([[1], []], r)
        [1]
        sage: from_fast_pickable([[0], []], r)
        [0]
        sage: from_fast_pickable([[2], [(0, 1, 0)]], r)
        [x(0)]
        sage: from_fast_pickable([[2], [(1, 1, 0)]], r)
        [x(1)]
        sage: from_fast_pickable([[2], [(0, 1, 1)]], r)
        [x(0) + 1]
        sage: from_fast_pickable([[2], [(0, 3, 0), (1, 1, 0)]], r)
        [x(0)*x(1)]
        sage: from_fast_pickable([[2], [(0, 3, 3), (1, 1, 0)]], r)
        [x(0)*x(1) + x(1)]
        sage: from_fast_pickable([[2], [(0, 3, 4), (1, 1, 0), (2, 1, 0)]], r)
        [x(0)*x(1) + x(2)]
        sage: from_fast_pickable([[2, 0, 1, 4], [(0, 3, 0), (1, 1, 0), (3, 1, 0)]], r)
        [x(0)*x(1), 0, 1, x(3)]
    """
    i2poly = {0: r.zero(), 1: r.one()}
    indices, terms = l

    for i in range(len(terms) - 1, -1, - 1):
        v, t, e = terms[i]
        t = i2poly[t]
        e = i2poly[e]
        terms[i] = if_then_else(v, t, e)
        i2poly[i + 2] = terms[i]
    return [Polynomial(i2poly[i]) for i in indices]


def _calculate_gb_with_keywords(args):
    I, kwds_as_single_arg = args
    return groebner_basis(I, **kwds_as_single_arg)


def _decode_polynomial(code):
    return from_fast_pickable(*code)[0]


def _encode_polynomial(poly):
    return (to_fast_pickable([poly]), poly.ring())


def pickle_polynomial(self):
    return (_decode_polynomial, (_encode_polynomial(self), ))


copyreg.pickle(Polynomial, pickle_polynomial)


def pickle_bset(self):
    return (BooleSet, (Polynomial(self), ))


copyreg.pickle(BooleSet, pickle_bset)


def pickle_monom(self):
    return (Monomial, (list(self.variables()),))


copyreg.pickle(Monomial, pickle_monom)


def pickle_var(self):
    return (Variable, (self.index(), self.ring()))


copyreg.pickle(Variable, pickle_var)


def _decode_ring(code):
    identifier, data, varnames, blocks = code

    global _polybori_parallel_rings
    try:
        _polybori_parallel_rings
    except NameError:
        _polybori_parallel_rings = {}

    for key in list(_polybori_parallel_rings):
        if not _polybori_parallel_rings[key][0]():
            del _polybori_parallel_rings[key]

    if identifier in _polybori_parallel_rings:
        ring = _polybori_parallel_rings[identifier][0]()
    else:
        ring = None

    if not ring:
        varnames = decompress(varnames).split('\n')
        nvars, ordercode = data
        ring = Ring(nvars, ordercode, names=varnames, blocks=blocks)
        storage_data = (WeakRingRef(ring), code)
        _polybori_parallel_rings[identifier] = storage_data
        _polybori_parallel_rings[(ring.id(), os.getpid())] = storage_data

    return ring


def _encode_ring(ring):
    identifier = (ring.id(), os.getpid())

    global _polybori_parallel_rings
    try:
        _polybori_parallel_rings
    except NameError:
        _polybori_parallel_rings = {}

    for key in list(_polybori_parallel_rings):
        if not _polybori_parallel_rings[key][0]():
            del _polybori_parallel_rings[key]

    if identifier in _polybori_parallel_rings:
        code = _polybori_parallel_rings[identifier][1]
    else:
        nvars = ring.n_variables()
        data = (nvars, ring.get_order_code())
        varnames = '\n'.join(str(ring.variable(idx))
                             for idx in range(nvars))
        blocks = list(ring.blocks())
        code = (identifier, data, compress(varnames), blocks[:-1])
        _polybori_parallel_rings[identifier] = (WeakRingRef(ring), code)

    return code


def pickle_ring(self):
    return (_decode_ring, (_encode_ring(self), ))


copyreg.pickle(Ring, pickle_ring)


def groebner_basis_first_finished(I, *l):
    r"""

    INPUT:

    - ``I`` -- ideal
    - ``l`` -- keyword dictionaries, which will be keyword arguments to
      ``groebner_basis``

    OUTPUT:

    - tries to compute ``groebner_basis(I, **kwd)`` for kwd in l
    - returns the result of the first terminated computation

    EXAMPLES::

        sage: from sage.rings.polynomial.pbori.PyPolyBoRi import Ring
        sage: r = Ring(1000)
        sage: ideal = [r.variable(1)*r.variable(2)+r.variable(2)+r.variable(1)]
        sage: from sage.rings.polynomial.pbori.parallel import groebner_basis_first_finished
        sage: groebner_basis_first_finished(ideal, dict(heuristic=True), dict(heuristic=False))
        [x1, x2]
    """
    if not I:
        return []

    from multiprocessing import Pool

    pool = Pool(processes=len(l))
    it = pool.imap_unordered(_calculate_gb_with_keywords,
                             [(I, kwds) for kwds in l])
    res = next(it)

    pool.terminate()

    return res



# sage_setup: distribution = sagemath-brial
# sage.doctest: needs sage.rings.polynomial.pbori
from sage.rings.polynomial.pbori.PyPolyBoRi import Polynomial, gauss_on_polys


def dense_system(I):
    I = (Polynomial(p) for p in I)
    I = (p for p in I if not p.is_zero())
    for p in I:
        d = p.deg()
        if d == 1:
            continue
        try:
            if len(p) > 2**d + 5:
                return True
        except OverflowError:
            return True
    return False


def gauss_on_linear(I):
    I = (Polynomial(p) for p in I)
    linear = []
    non_linear = []
    for p in I:
        if p.is_zero():
            continue
        if p.deg() <= 1:
            linear.append(p)
        else:
            non_linear.append(p)
    if not linear:
        return non_linear
    linear = list(gauss_on_polys(linear))
    return linear + non_linear



# sage_setup: distribution = sagemath-brial
# sage.doctest: needs sage.rings.polynomial.pbori
from sage.rings.polynomial.pbori.interpolate import (
    nf_lex_points,
    variety_lex_leading_terms,
)
from sage.rings.polynomial.pbori.pbori import easy_linear_factors


def easy_linear_polynomials(p):
    r"""
    Get linear polynomials implied by given polynomial.

    EXAMPLES::

        sage: from sage.rings.polynomial.pbori.frontend import x
        sage: from sage.rings.polynomial.pbori.easy_polynomials import easy_linear_polynomials
        sage: easy_linear_polynomials(x(1)*x(2) + 1)
        [x(1) + 1, x(2) + 1]
        sage: easy_linear_polynomials(x(1)*x(2) + 0)
        []
        sage: easy_linear_polynomials(x(0)*x(1) + x(0)*x(2) + 1)
        [x(0) + 1, x(1) + x(2) + 1]
    """
    res = []
    if p.deg() >= 2:
        if p.vars_as_monomial().deg() > 8:
            res.extend(q + 1 for q in easy_linear_factors(p + 1))
        else:
            res = easy_linear_polynomials_via_interpolation(p)
    return res


def easy_linear_polynomials_via_interpolation(p):
    r"""
    Get linear polynomials implied by given polynomial using interpolation of the variety.

    TESTS::

        sage: from sage.rings.polynomial.pbori.frontend import x
        sage: from sage.rings.polynomial.pbori.easy_polynomials import easy_linear_polynomials_via_interpolation
        sage: easy_linear_polynomials_via_interpolation(x(1)*x(2) + 1)
        [x(1) + 1, x(2) + 1]
        sage: easy_linear_polynomials_via_interpolation(x(1)*x(2) + 0)
        []
        sage: easy_linear_polynomials_via_interpolation(x(0)*x(1) + x(0)*x(2) + 1)
        [x(0) + 1, x(1) + x(2) + 1]
    """
    res = []
    p_vars = p.vars_as_monomial()
    space = p_vars.divisors()
    zeros = p.zeros_in(space)
    lex_leads = variety_lex_leading_terms(zeros, p_vars)
    for m in lex_leads:
        if m.deg() == 1:
            red = m + nf_lex_points(m, zeros)
            if red.lead_deg() == 1:  # normal ordering
                res.append(red)
    return res
